@@ -1,17 +1,17 @@
-"""tapdance.paths - Helper functions for dealing with paths."""
+"""tapdance.config - Helper functions for dealing with paths."""
 
 import json
 import os
 from pathlib import Path
 
 from logless import get_logger, logged
-from typing import Optional, Dict, Any
+from typing import Optional, Dict, Any, Tuple, List
 import uio
 
 logging = get_logger("tapdance")
 
 # These plugins will attempt to scrape and pass along AWs Credentials from the local environment.
-S3_TARGET_IDS = ["S3-CSV", "REDSHIFT", "SNOWFLAKE"]
+AWS_PLUGIN_IDS = ["S3-CSV", "REDSHIFT", "SNOWFLAKE", "DYNAMODB"]
 
 SINGER_PLUGINS_INDEX = os.environ.get("SINGER_PLUGINS_INDEX", "./singer_index.yml")
 VENV_ROOT = "/venv"
@@ -27,57 +27,121 @@ ENV_TAP_CONFIG_DIR = "TAP_CONFIG_DIR"
 ENV_TAP_STATE_FILE = "TAP_STATE_FILE"
 
 
+def validate_replication_strategy(replication_strategy):
+    if replication_strategy not in ["FULL_TABLE", "INCREMENTAL", "LOG_BASED"]:
+        raise ValueError(
+            f"Replication strategy {replication_strategy} not supported. Expected: "
+            "FULL_TABLE, INCREMENTAL, or LOG_BASED"
+        )
+
+
+def get_plugin_settings_from_env(plugin_name: str, meta_args: bool) -> Dict[str, Any]:
+    """Get all the settings from env vars which match the plugin prefix.
+
+    Parameters
+    ----------
+    plugin_name : str
+        The name of the plugin, including the 'tap-' or 'target-' prefix
+    meta_args : bool
+        True to pull _uppercase_ settings only, which are reserved for the orchestrator;
+        otherwise pull only _lowercase_ settings. By default false
+
+    Returns
+    -------
+    Dict[str, Any]
+        A dictionary of setting values.
+    """
+    conf_dict: Dict[str, Any] = {}
+    for k, v in os.environ.items():
+        prefix = f"{plugin_name.replace('-', '_').upper()}_"
+        if k.startswith(prefix):
+            logging.debug(f"Parsing env variable '{k}' for '{plugin_name}'...")
+            setting_name = k.split(prefix)[1]
+            case_match = meta_args == (setting_name.upper() == setting_name)
+            if case_match:
+                # Ensure truthinesss and falseness are maintained
+                if str(v).lower() == "false":
+                    conf_dict[setting_name] = False
+                elif str(v).lower() == "true":
+                    conf_dict[setting_name] = True
+                elif str(v) == "0":
+                    conf_dict[setting_name] = 0
+                else:
+                    conf_dict[setting_name] = v
+    return conf_dict
+
+
+def print_version():
+    """Print the tapdance version number."""
+    try:
+        from importlib import metadata
+    except ImportError:
+        # Running on pre-3.8 Python; use importlib-metadata package
+        import importlib_metadata as metadata
+    try:
+        version = metadata.version("tapdance")
+    except metadata.PackageNotFoundError:
+        version = "[could not be detected]"
+    print(f"tapdance version {version}")
+
+
 @logged(
     "getting '{plugin_name}' config file using: config_dir={config_dir}, "
-    "config_file={config_file}, required={required}",
+    "config_file={config_file}",
     log_fn=logging.debug,
 )
-def get_config_file(
+def get_or_create_config(
     plugin_name: str,
     taps_dir: str = None,
     config_dir: str = None,
     config_file: str = None,
-    required: bool = True,
-) -> str:
+) -> Tuple[str, Dict[str, Any]]:
     """
-    Return a path to the configuration file which also contains secrets.
+    Return a path to the configuration file and a dictionary of settings values.
 
-     - If file is blank or does not exist at the default secrets path, a new file will be created.
-     - If any environment variables exist in the form of TAP_MY_TAP_my_setting, a new file
-    will be created which contains these settings.
-     - If the default file exists and environment variables also exist, the temp file will
-    contain the default file values along with the environment variable overrides.
+     - If file is blank or does not exist at the default secrets path, a new file will
+       be created.
+     - If any environment variables exist in the form of TAP_MY_TAP_my_setting, a new
+       file will be created which contains these settings.
+     - If the default file exists and environment variables also exist, the temp file
+       will contain the default file values along with the environment variable
+       overrides.
+     - Unless `config_file=False` (by boolean or case-insensitive string comparison),
+       the config file must exist at the path provided or (if 'None' is passed) at the
+       default location.
     """
-    secrets_path = os.path.abspath(config_dir or get_secrets_dir(taps_dir))
-    config_file = config_file or f"{secrets_path}/{plugin_name}-config.json"
-    tmp_path = f"{secrets_path}/tmp/{plugin_name}-config.json"
     use_tmp_file = False
-    if uio.file_exists(config_file):
-        conf_dict = json.loads(uio.get_text_file_contents(config_file))
-    elif required:
-        raise FileExistsError(config_file)
-    else:
-        logging.info(f"No {plugin_name} config file exists. A file will be created...")
-        conf_dict = {}
+    secrets_path = os.path.abspath(config_dir or get_secrets_dir(taps_dir))
+    tmp_path = f"{secrets_path}/tmp/{plugin_name}-config.json"
+
+    orchestrator_env_vars = get_plugin_settings_from_env(plugin_name, meta_args=True)
+    config_file = config_file or orchestrator_env_vars.get("CONFIG_FILE", None)
+    if (config_file is not None) and str(config_file).lower() == "false":
+        logging.info(f"Skipping check for '{plugin_name}' config (`config_file=False`)")
         use_tmp_file = True
+        config_file = tmp_path
+        conf_dict = {}  # Start with empty config
+        orchestrator_settings = orchestrator_env_vars
+    else:
+        config_file = config_file or f"{secrets_path}/{plugin_name}-config.json"
+        if not uio.file_exists(config_file):
+            raise FileExistsError(
+                f"Could not find '{plugin_name}' config at expected path: {config_file}"
+            )
+        logging.info(f"Reading '{plugin_name}' config from {config_file}...")
+        conf_dict = json.loads(uio.get_text_file_contents(config_file))
+        orchestrator_settings = {
+            k: v
+            for k, v in conf_dict.items()
+            if k.upper() == k  # Uppercase settings only
+        }
+        orchestrator_settings.update(orchestrator_env_vars)
 
     # Parse settings and secrets from environment variables
-    for k, v in os.environ.items():
-        prefix = f"{plugin_name.replace('-', '_').upper()}_"
-        if k.startswith(prefix) and not k.endswith("_EXE"):
-            logging.debug(f"Parsing env variable '{k}' for '{plugin_name}'...")
-            setting_name = k.split(prefix)[1]
-            # Ensure truthinesss and falseness are maintained
-            if str(v).lower() == "false":
-                conf_dict[setting_name] = False
-            elif str(v).lower() == "true":
-                conf_dict[setting_name] = True
-            elif str(v) == "0":
-                conf_dict[setting_name] = 0
-            else:
-                conf_dict[setting_name] = v
-            use_tmp_file = True
-    if "-".join(plugin_name.split("-")[1:]).upper() in S3_TARGET_IDS:
+    conf_dict.update(get_plugin_settings_from_env(plugin_name, meta_args=False))
+    if conf_dict:
+        use_tmp_file = True
+    if "-".join(plugin_name.split("-")[1:]).upper() in AWS_PLUGIN_IDS:
         conf_dict = _inject_s3_config_creds(plugin_name, conf_dict)
         use_tmp_file = True
 
@@ -88,36 +152,13 @@ def get_config_file(
         config_file = tmp_path
     if not uio.file_exists(config_file):
         raise FileExistsError(config_file)
-    return config_file
+    return config_file, orchestrator_settings
 
 
 def get_pipeline_version_number() -> str:
     return os.environ.get(ENV_PIPELINE_VERSION_NUMBER, "1")
 
 
-def get_exe(tap_or_target_id: str) -> str:
-    """Get the exe for the tap or target.
-
-    Parameters
-    ----------
-    tap_or_target_id : str
-        The tap or target ID, including the tap-/target- prefix.
-
-    Returns
-    -------
-    str
-        The exe to use.
-    """
-    env_var = f"{tap_or_target_id.upper().replace('-', '_')}_EXE"
-    if env_var in os.environ:
-        result = os.environ[env_var]
-        logging.info(f"Found exe '{result}' for '{tap_or_target_id}' in env variable.")
-    else:
-        result = tap_or_target_id
-        logging.info(f"Defaulting to exe='{tap_or_target_id}' from tap name.")
-    return result
-
-
 def get_state_file_path(required: bool = True) -> Optional[str]:
     """Return a path to the state file or None if no state file path is configured.
 
@@ -192,6 +233,19 @@ def get_catalog_output_dir(tap_name: str, taps_dir: str) -> str:
     return result
 
 
+def get_raw_catalog_file(catalog_dir: str, tap_name: str):
+    custom_catalog_path = f"{catalog_dir}/{tap_name}-catalog-custom.json"
+    catalog_file = f"{catalog_dir}/{tap_name}-catalog-raw.json"
+    if uio.file_exists(custom_catalog_path):
+        logging.info(f"Using custom catalog file: {custom_catalog_path}")
+        catalog_file = custom_catalog_path
+    elif uio.file_exists(catalog_file):
+        if uio.get_text_file_contents(catalog_file).strip() == "":
+            logging.info(f"Cleaning up empty catalog file: {catalog_file}")
+            uio.delete_file(catalog_file)
+    return catalog_file.replace("\\", "/")
+
+
 def get_rules_file(taps_dir: str, tap_name: str, required: bool = True):
     result = os.path.join(get_taps_dir(taps_dir), f"{tap_name}.rules.txt")
     if required and not uio.file_exists(result):
@@ -206,9 +260,9 @@ def _inject_s3_config_creds(
         "aws_access_key_id" in config_defaults
         and "aws_secret_access_key" in config_defaults
     ):
-        logging.info("AWS creds captured from '{plugin_name}' config")
+        logging.info(f"AWS creds captured from '{plugin_name}' config")
         return config_defaults
-    logging.info("Scanning for AWS creds for '{plugin_name}' config...")
+    logging.info(f"Scanning for AWS creds for '{plugin_name}' config...")
     new_config = config_defaults.copy()
     (
         aws_access_key_id,
@@ -278,3 +332,41 @@ def replace_placeholders(
                         search_key, replacement_value
                     )
     return new_config
+
+
+def _dockerize_path(localpath: str, container_volume_root="/home/local") -> str:
+    result = os.path.relpath(localpath).replace("\\", "/")
+    result = f"{container_volume_root}/{result}"
+    return result
+
+
+def dockerize_cli_args(arg_str: str, container_volume_root="/home/local") -> str:
+    """Return a string with all host paths converted to their container equivalents.
+
+    Parameters
+    ----------
+    arg_str : str
+        The cli arg string to convert
+    container_volume_root : str, optional
+        The container directory which is mapped to local working directory,
+        by default "/home/local"
+
+    Returns
+    -------
+    str
+        A string with host paths converted to container paths.
+    """
+    args = arg_str.split(" ")
+    newargs: List[str] = []
+    for arg in args:
+        if uio.file_exists(arg):
+            newargs.append(_dockerize_path(arg, container_volume_root))
+        elif "=" in arg:
+            left, right = arg.split("=")[0], "=".join(arg.split("=")[1:])
+            if uio.file_exists(right):
+                newargs.append(
+                    f"{left}={_dockerize_path(right, container_volume_root)}"
+                )
+        else:
+            newargs.append(arg)
+    return " ".join(newargs)
