@@ -23,6 +23,7 @@ SKIP_SENSELESS_VALIDATORS = (
     # These only fail when our source is internally incoherent, which alone
     # is almost never sufficient cause for failing the data extraction.
 )
+SMALL_TABLE_THRESHOLD = 100000  # Warn if >= 10,000 rows and no replication key
 
 
 def _is_valid_json(json_text: str):
@@ -42,25 +43,38 @@ def _discover(
     dockerized: bool,
     tap_exe: str,
 ) -> None:
-    catalog_file = f"{catalog_dir}/{tap_name}-catalog-raw.json".replace("\\", "/")
+    catalog_file = config.get_raw_catalog_file(catalog_dir, tap_name)
     uio.create_folder(catalog_dir)
     img = f"{docker.BASE_DOCKER_REPO}:{tap_exe}"
+    hide_cmd = False
     if dockerized:
         cdw = os.getcwd().replace("\\", "/")
-        config_file = os.path.relpath(config_file).replace("\\", "/")
-        config_file = f"/home/local/{config_file}"
+        tap_config = json.loads(uio.get_text_file_contents(config_file))
+        tap_docker_args = ""
+        # TODO: Replace with logic to parse from AWS_SHARED_CREDENTIALS_FILE env var:
+        for k in ["aws_access_key_id", "aws_secret_access_key", "aws_session_token"]:
+            if k in tap_config:
+                key = f"TAP_{tap_name}_{k}".replace("-", "_").upper()
+                os.environ[key] = tap_config[k]
+                tap_docker_args += f' -e {k.upper()}="{tap_config[k]}"'
+                hide_cmd = True
+        _, _ = runnow.run(f"docker pull {img}")
         _, output_text = runnow.run(
-            f"docker run --rm -it "
-            f"-v {cdw}:/home/local "
-            f"{img} --config {config_file} --discover",
+            f"docker run --rm "
+            f"-v {cdw}:/home/local {tap_docker_args} "
+            f"{img} --config {config.dockerize_cli_args(config_file)} --discover",
             echo=False,
             capture_stderr=False,
+            hide=hide_cmd,
         )
         if not _is_valid_json(output_text):
             raise RuntimeError(f"Could not parse json file from output:\n{output_text}")
         uio.create_text_file(catalog_file, output_text)
     else:
-        runnow.run(f"{tap_exe} --config {config_file} --discover > {catalog_file}")
+        runnow.run(
+            f"{tap_exe} --config {config_file} --discover > {catalog_file}",
+            hide=hide_cmd,
+        )
 
 
 def _check_rules(
@@ -87,6 +101,13 @@ def _check_rules(
         for line in uio.get_text_file_contents(rules_file).splitlines()
         if line.split("#")[0].rstrip()
     ]
+    declared_tables = set(
+        [
+            rule.split(".")[0].rstrip().lstrip("!")
+            for rule in select_rules
+            if rule.split(".")[0].rstrip() and ("*" not in rule.split(".")[0])
+        ]
+    )
     matches: Dict[str, dict] = {}
     excluded_table_list = []
     for table_name, table_object in _get_catalog_tables_dict(catalog_file).items():
@@ -101,6 +122,18 @@ def _check_rules(
                 )
         else:
             excluded_table_list.append(table_name)
+    for required_table in declared_tables:
+        if required_table not in matches.keys():
+            logging.warning(
+                f"The table {required_table} was declared in the rules file "
+                "but could not be found in the catalog."
+            )
+    for match, match_cols in matches.items():
+        if not match_cols:
+            logging.warning(
+                f"The table {match} was declared in the rules file "
+                "but did not match with any columns in the catalog."
+            )
     return matches, excluded_table_list
 
 
@@ -108,20 +141,39 @@ def _validate_keys(table_object: dict, key_type: str):
     _ = _get_table_key_cols(key_type, table_object, warn_if_missing=True)
 
 
-def _get_table_key_cols(key_type: str, table_object: dict, warn_if_missing: bool):
+def _get_table_key_cols(
+    key_type: str, table_object: dict, warn_if_missing: bool, log_fn=None
+):
     result = []
     metadata_object = _get_stream_metadata_object(table_object)
+    row_count: Optional[int] = metadata_object.get("row-count", None)
+    row_count_desc = f" (est. {row_count:,} rows)" if row_count else ""
+    view_warning = f"(is-view=TRUE)" if metadata_object.get("is-view", None) else ""
     if key_type == "replication-key":
+        if not log_fn:
+            if row_count and row_count < SMALL_TABLE_THRESHOLD:
+                log_fn = logging.debug
+            else:
+                log_fn = logging.warning
         if "valid-replication-keys" in metadata_object:
             result = metadata_object["valid-replication-keys"]
     elif key_type == "primary-key":
+        log_fn = log_fn or logging.warning
         if "table-key-properties" in metadata_object:
             result = metadata_object["table-key-properties"]
     else:
         raise ValueError("Expected key_type of 'primary-key' or 'replication-key'")
     if not result and warn_if_missing:
         table_name = table_object.get("stream", "(unknown stream)")
-        logging.warning(f"Could not locate '{key_type}' for '{table_name}'.")
+        log_fn(
+            " ".join(
+                [
+                    f"Could not locate '{key_type}' for '{table_name}'",
+                    row_count_desc,
+                    view_warning,
+                ]
+            )
+        )
     return result
 
 
@@ -156,6 +208,7 @@ def _set_catalog_file_keys(table_object: dict, table_plan: dict):
                 + ")"
             )
             metadata["valid-replication-keys"] = table_plan["replication_key"]
+            metadata["replication-key"] = table_plan["replication_key"][0]
 
 
 def _get_catalog_file_keys(
@@ -232,6 +285,23 @@ def get_table_list(
     exclude_tables: Optional[Union[str, List[str]]],
     catalog_file: str,
 ) -> List[str]:
+    """Return a list of tables to be selected.
+
+    Parameters
+    ----------
+    table_filter : Optional[Union[str, List[str]]]
+        A list of tables, a string representing a list, a single table name, or "*"
+        for all tables.
+    exclude_tables : Optional[Union[str, List[str]]]
+        A list of tables or a string representing a list of tables.
+    catalog_file : str
+        The path to a catalog file.
+
+    Returns
+    -------
+    List[str]
+        A list of table names selected.
+    """
     if isinstance(table_filter, list):
         list_of_tables = table_filter
     elif table_filter is None or table_filter == "*":
@@ -242,22 +312,26 @@ def get_table_list(
     else:
         list_of_tables = [table_filter]
     if exclude_tables:
+        if exclude_tables[0] == "[":
+            # Remove square brackets and split the result on commas
+            exclude_tables = (
+                str(exclude_tables).replace("[", "").replace("]", "").split(",")
+            )
         logging.info(f"Table(s) to exclude from sync: {', '.join(exclude_tables)}")
         list_of_tables = [t for t in list_of_tables if t not in exclude_tables]
     return list_of_tables
 
 
-@logged("updating plan file for 'tap-{tap_name}'")
 def plan(
     tap_name: str,
     *,
+    dockerized: bool = None,
     rescan: bool = None,
+    tap_exe: str = None,
     taps_dir: str = None,
     config_dir: str = None,
     config_file: str = None,
-    dockerized: bool = None,
-    tap_exe: str = None,
-    replication_strategy: str = "INCREMENTAL",
+    replication_strategy: str = None,
 ) -> None:
     """Perform all actions necessary to prepare (plan) for a tap execution.
 
@@ -271,27 +345,27 @@ def plan(
 
     Parameters:
     -----------
-    tap_name : str
+    tap_name : {str}
         The name of the tap without the 'tap-' prefix.
-    rescan : bool, optional
-        True to force a rescan and replace existing metadata. (default: False)
-    taps_dir: str, optional
+    dockerized : {bool}
+        If specified, will override the default behavior for the local platform.
+    tap_exe : {str}
+        Specifies the tap executable, if different from `tap-{tap_name}`.
+    rescan : {bool}
+        True to force a rescan and replace existing metadata.
+    taps_dir: {str}
         The directory containing the rules file.
         (Default=cwd)
-        (e.g. `{tap-name}.rules.txt`).
-    config_dir: str, optional
+    config_dir: {str}
         The default location of config, catalog and other potentially sensitive
         information. (Recommended to be excluded from source control.)
         (Default="${taps_dir}/.secrets")
-    config_file : str, optional
+    config_file : {str}
         The location of the JSON config file which contains config for the specified
         plugin. (Default=f"${config_dir}/${plugin_name}-config.json")
-    dockerized : bool, optional
-        If specified, will override the default behavior for the local platform.
-    tap_exe : str, optional
-        Specifies the tap executable, if different from `tap-{tap_name}`.
-    replication_strategy : str, optional
-        One of "FULL_TABLE", "INCREMENTAL", or "LOG_BASED"; by default "INCREMENTAL"
+    replication_strategy : {str}
+        One of "FULL_TABLE", "INCREMENTAL", or "LOG_BASED"; by default "INCREMENTAL" or
+        a value is set in the TAP_{TAPNAME}_REPLICATION_STRATEGY environment variable.
 
     Raises
     ------
@@ -301,46 +375,36 @@ def plan(
         Raised if files do not exist in default locations, or if paths provided do not
         point to valid files.
     """
-    if replication_strategy not in ["FULL_TABLE", "INCREMENTAL", "LOG_BASED"]:
-        raise ValueError(
-            f"Replication strategy {replication_strategy} not supported. Expected: "
-            "FULL_TABLE, INCREMENTAL, or LOG_BASED"
-        )
-    tap_exe = tap_exe or config.get_exe(f"tap-{tap_name}")
-    if (dockerized is None) and (uio.is_windows() or uio.is_mac()):
-        dockerized = True
-        logging.info(
-            "The 'dockerized' argument is not set when running either Windows or OSX. "
-            "Defaulting to dockerized=True..."
-        )
+    config.print_version()
 
-    # Initialize paths
     taps_dir = config.get_taps_dir(taps_dir)
-    config_required = True
-    if (config_file is not None) and str(config_file).lower() == "false":
-        config_file = None
-        config_required = False
-    config_file = config.get_config_file(
+    config_file, tap_settings = config.get_or_create_config(
         f"tap-{tap_name}",
         taps_dir=taps_dir,
         config_dir=config_dir,
         config_file=config_file,
-        required=config_required,
     )
-    if not uio.file_exists(config_file):
-        raise FileExistsError(config_file)
+    tap_exe = tap_exe or tap_settings.get("EXE", f"tap-{tap_name}")
+    replication_strategy = replication_strategy or tap_settings.get(
+        "REPLICATION_STRATEGY", "INCREMENTAL"
+    )
+    config.validate_replication_strategy(replication_strategy)
+
+    # TODO: Resolve bug in Windows STDERR inclusion when emitting catalog json from
+    #       docker run
+    # if (dockerized is None) and (uio.is_windows() or uio.is_mac()):
+    #     dockerized = True
+    #     logging.info(
+    #         "The 'dockerized' argument is not set when running either Windows or OSX. "
+    #         "Defaulting to dockerized=True..."
+    #     )
+
     catalog_dir = config.get_catalog_output_dir(tap_name, taps_dir)
-    catalog_file = f"{catalog_dir}/{tap_name}-catalog-raw.json"
+    raw_catalog_file = config.get_raw_catalog_file(catalog_dir, tap_name)
     selected_catalog_file = f"{catalog_dir}/{tap_name}-catalog-selected.json"
     plan_file = config.get_plan_file(tap_name, taps_dir, required=False)
-    if (
-        uio.file_exists(catalog_file)
-        and uio.get_text_file_contents(catalog_file).strip() == ""
-    ):
-        logging.info(f"Cleaning up empty catalog file: {catalog_file}")
-        uio.delete_file(catalog_file)
 
-    if rescan or not uio.file_exists(catalog_file):
+    if rescan or not uio.file_exists(raw_catalog_file):
         # Run discover, if needed, to get catalog.json (raw)
         _discover(
             tap_name,
@@ -352,16 +416,16 @@ def plan(
 
     rules_file = config.get_rules_file(taps_dir, tap_name)
     matches, excluded_tables = _check_rules(
-        catalog_file=catalog_file, rules_file=rules_file
+        catalog_file=raw_catalog_file, rules_file=rules_file
     )
     primary_keys = _get_catalog_file_keys(
-        "primary-key", matches=matches, catalog_file=catalog_file
+        "primary-key", matches=matches, catalog_file=raw_catalog_file
     )
     primary_keys.update(
         _get_rules_file_keys("primary-key", matches=matches, rules_file=rules_file)
     )
     replication_keys = _get_catalog_file_keys(
-        "replication-key", matches=matches, catalog_file=catalog_file,
+        "replication-key", matches=matches, catalog_file=raw_catalog_file,
     )
     replication_keys.update(
         _get_rules_file_keys("replication-key", matches=matches, rules_file=rules_file,)
@@ -373,7 +437,7 @@ def plan(
     _create_selected_catalog(
         tap_name,
         plan_file=plan_file,
-        full_catalog_file=catalog_file,
+        raw_catalog_file=raw_catalog_file,
         output_file=selected_catalog_file,
         replication_strategy=replication_strategy,
         skip_senseless_validators=SKIP_SENSELESS_VALIDATORS,
@@ -402,29 +466,42 @@ def _make_plan_file_text(
             and col not in replication_key_cols
         ]
         ignored_cols = [col for col, selected in matches[table].items() if not selected]
-        file_text += f"{'  ' * 1}{table}:\n"
-        file_text += f"{'  ' * 2}primary_key:\n"
+        file_text += f"{' ' * 2}{table}:\n"
+        file_text += f"{' ' * 4}primary_key:\n"
         for col in primary_key_cols:
-            file_text += f"{'  ' * 2}- {col}\n"
-        file_text += f"{'  ' * 2}replication_key:\n"
+            file_text += f"{' ' * 6}- {col}\n"
+        file_text += f"{' ' * 4}replication_key:\n"
         for col in replication_key_cols:
-            file_text += f"{'  ' * 2}- {col}\n"
-        file_text += f"{'  ' * 2}selected_columns:\n"
+            file_text += f"{' ' * 6}- {col}\n"
+        file_text += f"{' ' * 4}selected_columns:\n"
         for col in included_cols:
-            file_text += f"{'  ' * 2}- {col}\n"
+            file_text += f"{' ' * 6}- {col}\n"
         if ignored_cols:
-            file_text += f"{'  ' * 2}ignored_columns:\n"
+            file_text += f"{' ' * 4}ignored_columns:\n"
             for col in ignored_cols:
-                file_text += f"{'  ' * 2}- {col}\n"
+                file_text += f"{' ' * 6}- {col}\n"
     if excluded_tables_list:
         file_text += "ignored_tables:\n"
         for table in sorted(excluded_tables_list):
-            file_text += f"{'  ' * 1}- {table}\n"
+            file_text += f"{' ' * 2}- {table}\n"
     return file_text
 
 
 def _get_catalog_table_columns(table_object: dict) -> List[str]:
-    return table_object["schema"]["properties"].keys()
+    col_names: List[str] = []
+    if "properties" not in table_object["schema"]:
+        if table_object["schema"].get("type", "") == "object":
+            logging.warning(
+                "No properties defined, treating record as single json_raw column."
+            )
+            col_names = ["json_raw"]
+    else:
+        col_names = table_object["schema"]["properties"].keys()
+    if not col_names:
+        raise ValueError(
+            f"Could not detect table columns from catalog table: {table_object}"
+        )
+    return col_names
 
 
 def _get_catalog_tables_dict(catalog_file: str) -> dict:
@@ -433,29 +510,32 @@ def _get_catalog_tables_dict(catalog_file: str) -> dict:
     return table_objects
 
 
-def _get_stream_name(table_object: dict) -> dict:
+def _get_stream_name(table_object: dict) -> str:
     return table_object["stream"]
 
 
+def _get_table_name(table_object: dict) -> str:
+    return table_object.get(
+        "table_name", table_object.get("stream", table_object["tap_stream_id"])
+    )
+
+
 @logged(
     "selecting catalog metadata "
-    "from '{tap_name}' source catalog file: {full_catalog_file}"
+    "from '{tap_name}' source catalog file: {raw_catalog_file}"
 )
 def _create_selected_catalog(
     tap_name: str,
     plan_file: str,
-    full_catalog_file: str,
+    raw_catalog_file: str,
     output_file: str,
     replication_strategy: str,
     skip_senseless_validators: bool,
 ) -> None:
     taps_dir = config.get_taps_dir()
     catalog_dir = config.get_catalog_output_dir(tap_name, taps_dir)
-    source_catalog_path = full_catalog_file or os.path.join(
-        catalog_dir, "catalog-raw.json"
-    )
     output_file = output_file or os.path.join(catalog_dir, "selected-catalog.json")
-    catalog_full = json.loads(Path(source_catalog_path).read_text())
+    catalog_full = json.loads(Path(raw_catalog_file).read_text())
     plan_file = plan_file or config.get_plan_file(tap_name)
     plan = yaml.safe_load(uio.get_text_file_contents(plan_file))
     if ("selected_tables" not in plan) or (plan["selected_tables"] is None):
@@ -464,11 +544,13 @@ def _create_selected_catalog(
     for tbl in sorted(catalog_full["streams"], key=lambda x: _get_stream_name(x)):
         stream_name = _get_stream_name(tbl)
         if stream_name in plan["selected_tables"].keys():
-            _select_table(tbl, replication_strategy=replication_strategy)
             _set_catalog_file_keys(tbl, plan["selected_tables"][stream_name])
+            _select_table(tbl, replication_strategy=replication_strategy)
             for col_name in _get_catalog_table_columns(tbl):
-                col_selected = (
-                    col_name in plan["selected_tables"][stream_name]["selected_columns"]
+                col_selected = col_name in (
+                    (plan["selected_tables"][stream_name]["selected_columns"] or [])
+                    + (plan["selected_tables"][stream_name]["replication_key"] or [])
+                    + (plan["selected_tables"][stream_name]["primary_key"] or [])
                 )
                 _select_table_column(tbl, col_name, col_selected)
             if skip_senseless_validators:
@@ -519,6 +601,8 @@ def _select_table(tbl: dict, replication_strategy: str):
 
 
 def _remove_senseless_validators(tbl: dict) -> None:
+    if "properties" not in tbl["schema"]:
+        return
     for col, props in tbl["schema"]["properties"].items():
         for senseless in [
             "multipleOf",
